from .tts import (
    TTS,
    SynthesisEvent,
    SynthesizedAudio,
    SynthesizeStream,
    SynthesisEventType,
)

from .stream_adapter import StreamAdapterWrapper

__all__ = [
    "TTS",
    "SynthesisEvent",
    "SynthesizedAudio",
    "SynthesizeStream",
    "SynthesisEventType",
    "StreamAdapterWrapper",
]



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .version import __version__

from .worker import (
    Worker,
    JobCancelledError,
    AssignmentTimeoutError,
    JobType,
    run_app,
)
from .plugin import Plugin
from .utils import AudioBuffer, merge_frames
from .job_request import AutoSubscribe, AutoDisconnect, JobRequest
from .job_context import JobContext

from . import stt
from . import vad
from . import tts
from . import tokenize

__all__ = [
    "__version__",
    "Worker",
    "JobRequest",
    "AutoSubscribe",
    "AutoDisconnect",
    "JobContext",
    "JobCancelledError",
    "AssignmentTimeoutError",
    "Plugin",
    "run_app",
    "JobType",
    "AudioBuffer",
    "merge_frames",
    "stt",
    "vad",
    "tts",
    "tokenize",
]



from abc import ABC, abstractmethod
from typing import List
from dataclasses import dataclass, field
from livekit import rtc
from enum import Enum


class VADEventType(Enum):
    START_SPEAKING = 1
    SPEAKING = 2
    END_SPEAKING = 3


@dataclass
class VADEvent:
    # type of the event
    type: VADEventType
    # index of the samples of the event (when the event was fired)
    samples_index: int
    # duration of the speech in seconds (only for END_SPEAKING event)
    duration: float = 0.0
    # list of audio frames of the speech
    speech: List[rtc.AudioFrame] = field(default_factory=list)


class VAD(ABC):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def stream(
        self,
        *,
        min_speaking_duration: float = 0.5,
        min_silence_duration: float = 0.5,
        padding_duration: float = 0.1,
        sample_rate: int = 16000,
        max_buffered_speech: float = 45.0,
    ) -> "VADStream":
        """Returns a VADStream that can be used to push audio frames and receive VAD events
        Args:
          min_speaking_duration: minimum duration of speech to trigger a START_SPEAKING event

          min_silence_duration: in the end of each speech chunk wait for min_silence_duration_ms before separating it
              the padding is not always precise, it is generally rounded to the nearest 40ms depending on the vad implementation

          padding_duration: frames to pad the start and end of the speech with

          sample_rate: sample rate of the inference/processing

          max_buffered_speech: max buffered speech in seconds that we keep until the END_SPEAKING event is triggered
              it is unrecommended to use 0.0 as it may cause OOM if the user doesn't stop speaking"""
        pass


class VADStream(ABC):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def push_frame(self, frame: rtc.AudioFrame) -> None:
        pass

    @abstractmethod
    async def flush(self) -> None:
        pass

    @abstractmethod
    async def aclose(self) -> None:
        pass

    @abstractmethod
    async def __anext__(self) -> VADEvent:
        raise StopAsyncIteration

    def __aiter__(self) -> "VADStream":
        return self



from .stt import (
    SpeechEvent,
    SpeechData,
    SpeechStream,
    STT,
)
from .stream_adapter import StreamAdapter, StreamAdapterWrapper

__all__ = [
    "SpeechEvent",
    "SpeechData",
    "SpeechStream",
    "STT",
    "StreamAdapter",
    "StreamAdapterWrapper",
]



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from typing import Callable, Coroutine, Optional, TYPE_CHECKING, Union
from .job_context import JobContext
from livekit import rtc, protocol, api
from livekit.protocol import agent as proto_agent

# TODO: refactor worker so we can avoid this circular import
if TYPE_CHECKING:
    from .worker import Worker

AutoSubscribeCallback = Callable[[rtc.TrackPublication, rtc.RemoteParticipant], bool]
AutoDisconnectCallback = Callable[[JobContext], bool]
AgentEntry = Callable[[JobContext], Coroutine]


class AutoSubscribe:
    """Helper callbacks for common subscribe scenarios"""

    @staticmethod
    def SUBSCRIBE_ALL(
        publication: rtc.TrackPublication, participant: rtc.RemoteParticipant
    ) -> bool:
        """Subscribe to all tracks automatically. This will also set the LiveKit room option
        auto_subscribe to true as an optimization."""
        return True

    @staticmethod
    def SUBSCRIBE_NONE(
        publication: rtc.TrackPublication, participant: rtc.RemoteParticipant
    ) -> bool:
        """Don't subscribe to any tracks automatically"""
        return False

    @staticmethod
    def VIDEO_ONLY(
        publication: rtc.TrackPublication, participant: rtc.RemoteParticipant
    ) -> bool:
        """Subscribe to video tracks automatically"""
        return publication.kind == rtc.TrackKind.KIND_VIDEO

    @staticmethod
    def AUDIO_ONLY(
        publication: rtc.TrackPublication, participant: rtc.RemoteParticipant
    ) -> bool:
        """Subscribe to audio tracks automatically"""
        return publication.kind == rtc.TrackKind.KIND_AUDIO


class AutoDisconnect:
    """Helper callbacks for common auto disconnect scenarios"""

    @staticmethod
    def ROOM_EMPTY(ctx: JobContext) -> bool:
        if len(ctx.room.participants) == 0:
            return True

        if len(ctx.room.participants) > 1:
            return False

        for p in ctx.room.participants.values():
            if p.identity == ctx.agent.identity:
                return True

        return False

    @staticmethod
    def PUBLISHER_LEFT(ctx: JobContext) -> bool:
        if ctx.participant is None:
            logging.error(
                "Incorrect usage of PUBLISHER_LEFT, JobContext is tied to a Participant",
                extra=ctx.logging_extra,
            )
            return False

        return ctx.room.participants.get(ctx.participant.sid) is None

    @staticmethod
    def DEFAULT(ctx: JobContext):
        """
        Default shutdown options. If the agent is tied to a participant, it will shut down when that participant leaves.
        If the agent is not tied to a participant, it will shut down when the agent is the only remaining participant.
        """
        if ctx.participant is not None:
            return AutoDisconnect.PUBLISHER_LEFT(ctx)

        return AutoDisconnect.ROOM_EMPTY(ctx)


class JobRequest:
    """Represents a new job from the server, this worker can either accept or reject it."""

    def __init__(
        self,
        worker: "Worker",
        job_info: proto_agent.Job,
        simulated: bool = False,
    ) -> None:
        self._worker = worker
        self._info = job_info
        self._room = rtc.Room()
        self._answered = False
        self._simulated = simulated
        self._lock = asyncio.Lock()

    @property
    def id(self) -> str:
        return self._info.id

    @property
    def room(self) -> protocol.models.Room:
        return self._info.room

    @property
    def publisher(self) -> Optional[protocol.models.ParticipantInfo]:
        return self._info.participant

    async def reject(self) -> None:
        """Tell the server we cannot handle this job"""
        async with self._lock:
            if self._answered:
                raise Exception("job already answered")

            self._answered = True
            if not self._simulated:
                await self._worker._send_availability(self.id, False)

        logging.info("rejected job %s", self.id)

    async def accept(
        self,
        agent: AgentEntry,
        auto_subscribe: Union[
            AutoSubscribeCallback, bool
        ] = AutoSubscribe.SUBSCRIBE_NONE,
        auto_disconnect: AutoDisconnectCallback = AutoDisconnect.DEFAULT,
        grants: api.VideoGrants = api.VideoGrants(),
        name: Optional[str] = None,
        identity: Optional[str] = None,
        metadata: Optional[str] = None,
    ) -> None:
        """Signal to the LiveKit Server that we can handle the job, and connect to the room.

        Args:
            agent:
                Your agent entrypoint.

            auto_subscribe:
                Callback that is called when a track is published in the room. Return True to subscribe to the track.

            auto_disconnect:
                Callback that is called when the agent should disconnect from the room. Return True to disconnect.
                The callback is called when:
                - Initially once the agent has connected to the room
                - A participant leaves the room

            grants:
                Additional grants to give to the agent participant in its token.
                Defaults to None.

            name:
                Name of the agent participant. Defaults to "".

            identity:
                Identity of the agent participant. Defaults to "".

            metadata:
                Metadata of the agent participant. Defaults to "".
        """
        async with self._lock:
            if self._answered:
                raise Exception("job already answered")

            self._answered = True

            identity = identity or "agent-" + self.id
            grants.room_join = True
            grants.agent = True
            grants.room = self.room.name
            grants.can_update_own_metadata = True

            jwt = (
                api.AccessToken(self._worker._api_key, self._worker._api_secret)
                .with_identity(identity)
                .with_grants(grants)
                .with_metadata(metadata or "")
                .with_name(name or "")
                .to_jwt()
            )

            # raise AssignmentTimeoutError if assignment times out
            if not self._simulated:
                _ = await self._worker._send_availability(self.id, True)

            try:
                options = rtc.RoomOptions(
                    auto_subscribe=auto_subscribe == AutoSubscribe.SUBSCRIBE_ALL
                    or auto_subscribe is True
                )
                await self._room.connect(self._worker._rtc_url, jwt, options)
            except rtc.ConnectError as e:
                logging.exception(
                    "failed to connect to the room, cancelling job %s",
                    self.id,
                    extra={
                        "job_id": self.id,
                        "room": self.room.name,
                        "agent_identity": identity,
                    },
                )
                await self._worker._send_job_status(
                    self.id, proto_agent.JobStatus.JS_FAILED, str(e)
                )
                raise

            participant: Optional[rtc.Participant] = None
            if self._info.participant:
                participant = self._room.participants.get(self._info.participant.sid)

            job_ctx = JobContext(
                self.id,
                self._worker,
                self._room,
                participant=participant,
            )
            job_ctx.create_task(agent(job_ctx))

            def disconnect_if_needed(*_):
                if auto_disconnect(job_ctx):
                    asyncio.ensure_future(
                        job_ctx.disconnect(),
                        loop=self._worker._loop,
                    )

            self._room.on("participant_disconnected", disconnect_if_needed)

            async def job_validity():
                # let time to the server to forward participant info
                await asyncio.sleep(15)
                disconnect_if_needed()

            job_ctx.create_task(job_validity())

            if auto_subscribe != AutoSubscribe.SUBSCRIBE_ALL and not isinstance(
                auto_subscribe, bool
            ):

                @self._room.on("track_published")
                def on_track_published(
                    publication: rtc.RemoteTrackPublication,
                    participant: rtc.RemoteParticipant,
                ):
                    if not auto_subscribe(publication, participant):
                        return

                    publication.set_subscribed(True)

                for participant in self._room.participants.values():
                    for publication in participant.tracks.values():
                        if not auto_subscribe(publication, participant):
                            continue

                        publication.set_subscribed(True)

        logging.info("accepted job %s", self.id, extra=job_ctx.logging_extra)



import contextlib
import asyncio
import logging
from typing import Optional
from ..vad import VADStream, VADEventType
from ..utils import merge_frames, AudioBuffer
from .stt import (
    STT,
    SpeechStream,
    SpeechEvent,
)
from livekit import rtc


class StreamAdapter(STT):
    def __init__(
        self,
        stt: STT,
        vad_stream: VADStream,
    ) -> None:
        super().__init__(streaming_supported=True)
        self._vad_stream = vad_stream
        self._stt = stt

    async def recognize(self, *, buffer: AudioBuffer, language: Optional[str] = None):
        return await self._stt.recognize(
            buffer=buffer,
            language=language,
        )

    def stream(
        self,
        *,
        language: Optional[str] = None,
    ) -> SpeechStream:
        return StreamAdapterWrapper(
            self._vad_stream,
            self._stt,
            language=language,
        )


class StreamAdapterWrapper(SpeechStream):
    def __init__(
        self,
        vad_stream: VADStream,
        stt: STT,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self._vad_stream = vad_stream
        self._stt = stt
        self._event_queue = asyncio.Queue[SpeechEvent]()
        self._closed = False
        self._args = args
        self._kwargs = kwargs

        self._main_task = asyncio.create_task(self._run())

        def log_exception(task: asyncio.Task) -> None:
            if not task.cancelled() and task.exception():
                logging.error(f"stream adapter task failed: {task.exception()}")

        self._main_task.add_done_callback(log_exception)

    async def _run(self) -> None:
        # listen to vad events and send to stt on END_SPEAKING
        try:
            async for event in self._vad_stream:
                if event.type != VADEventType.END_SPEAKING:
                    continue

                merged_frames = merge_frames(event.speech)
                event = await self._stt.recognize(
                    buffer=merged_frames, *self._args, **self._kwargs
                )
                self._event_queue.put_nowait(event)
        except asyncio.CancelledError:
            pass

        self._closed = True

    def push_frame(self, frame: rtc.AudioFrame) -> None:
        self._vad_stream.push_frame(frame)

    async def aclose(self) -> None:
        await self._vad_stream.aclose()
        self._main_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self._main_task

    async def flush(self) -> None:
        await self._vad_stream.flush()

    async def __anext__(self) -> SpeechEvent:
        if self._closed and self._event_queue.empty():
            raise StopAsyncIteration

        return await self._event_queue.get()



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import logging
from typing import Coroutine, Optional, TYPE_CHECKING
from livekit import api, rtc, protocol

# TODO: refactor worker so we can avoid this circular import
if TYPE_CHECKING:
    from .worker import Worker


class JobContext:
    """Context for job, it contains the worker, the room, and the participant.
    You should not create these on your own. They are created by the Worker.
    """

    def __init__(
        self,
        id: str,
        worker: "Worker",
        room: rtc.Room,
        participant: Optional[rtc.Participant] = None,
    ) -> None:
        self._id = id
        self._worker = worker
        self._room = room
        self._participant = participant
        self._closed = False
        self._lock = asyncio.Lock()
        self._worker._running_jobs.append(self)
        self._tasks = set()

    @property
    def id(self) -> str:
        """Job ID"""
        return self._id

    @property
    def room(self) -> rtc.Room:
        """LiveKit Room object"""
        return self._room

    @property
    def participant(self) -> Optional[rtc.Participant]:
        """LiveKit RemoteParticipant that the Job launched for. None if Agent is launched for the Room."""
        return self._participant

    @property
    def agent(self) -> rtc.LocalParticipant:
        """The agent's Participant identity"""
        return self._room.local_participant

    @property
    def api(self) -> api.LiveKitAPI:
        """LiveKit API client"""
        return self._worker.api

    def create_task(self, coro: Coroutine) -> asyncio.Task:
        """Schedule the execution of a coroutine object in a spawn task using asyncio and keep a reference to it.
        The task is automatically cancelled when the job is disconnected.
        """

        t = asyncio.create_task(coro)

        def done_cb(task: asyncio.Task):
            self._tasks.discard(t)
            if not task.cancelled() and task.exception():
                logging.error(
                    "A task raised an exception:",
                    exc_info=task.exception(),
                    extra=self.logging_extra,
                )

        t.add_done_callback(done_cb)
        return t

    async def disconnect(self) -> None:
        """
        Disconnect the agent from the room, shutdown the job, and cleanup resources.
        This will also cancel all tasks created by this job if task_timeout is specified.

        Args:
            task_timeout (Optional[float], optional): How long to wait before tasks created via JobContext.create_task are cancelled.
            If None, tasks will not be cancelled. Defaults to 25.
        """

        async with self._lock:
            logging.info("shutting down job %s", self.id, extra=self.logging_extra)
            if self._closed:
                return

            self._worker._running_jobs.remove(self)
            self._closed = True
            await self.room.disconnect()

            for task in self._tasks:
                task.cancel()

            for task in self._tasks:
                with contextlib.suppress(asyncio.CancelledError):
                    await task

            logging.info("job %s shutdown", self.id, extra=self.logging_extra)

    async def update_status(
        self,
        status: protocol.agent.JobStatus.ValueType,
        error: str = "",
        user_data: str = "",
    ) -> None:
        await self._worker._send_job_status(self._id, status, error, user_data)

    @property
    def logging_extra(self) -> dict:
        """
        Additional context to identify the job in logs.

        Usage: logging.info("my message", extra=ctx.logging_extra)
        """
        e = {
            "job_id": self.id,
            "room": self.room.name,
            "agent_identity": self.agent.identity,
            "worker_id": self._worker.id,
        }
        if self.participant:
            e["participant_identity"] = self.participant.identity
        return e



from abc import ABC, abstractmethod
from livekit import rtc
from typing import Optional, List
from dataclasses import dataclass
from ..utils import AudioBuffer


@dataclass
class SpeechData:
    language: str
    text: str
    start_time: float = 0.0
    end_time: float = 0.0
    confidence: float = 0.0  # [0, 1]


@dataclass
class SpeechEvent:
    is_final: bool
    alternatives: List[SpeechData]


class STT(ABC):
    def __init__(
        self,
        *,
        streaming_supported: bool,
    ) -> None:
        self._streaming_supported = streaming_supported

    @abstractmethod
    async def recognize(
        self,
        *,
        buffer: AudioBuffer,
        language: Optional[str] = None,
    ) -> SpeechEvent:
        pass

    def stream(
        self,
        *,
        language: Optional[str] = None,
    ) -> "SpeechStream":
        raise NotImplementedError(
            "streaming is not supported by this STT, please use a different STT or use a StreamAdapter"
        )

    @property
    def streaming_supported(self) -> bool:
        return self._streaming_supported


class SpeechStream(ABC):
    @abstractmethod
    def push_frame(self, frame: rtc.AudioFrame) -> None:
        pass

    @abstractmethod
    async def flush(self) -> None:
        pass

    @abstractmethod
    async def aclose(self) -> None:
        pass

    @abstractmethod
    async def __anext__(self) -> SpeechEvent:
        pass

    def __aiter__(self) -> "SpeechStream":
        return self



from abc import ABC, abstractmethod
from typing import Optional
from enum import Enum
from livekit import rtc
from dataclasses import dataclass


@dataclass
class SynthesizedAudio:
    text: str
    data: rtc.AudioFrame


class SynthesisEventType(Enum):
    # first event, indicates that the stream has started
    # retriggered after FINISHED
    STARTED = 0
    # audio data is available
    AUDIO = 1
    # generally happens after a flushing the stream
    # some TTS providers close the stream so we know it's done
    FINISHED = 2


@dataclass
class SynthesisEvent:
    type: SynthesisEventType
    audio: Optional[SynthesizedAudio] = None


class SynthesizeStream(ABC):
    @abstractmethod
    def push_text(self, token: str) -> None:
        pass

    @abstractmethod
    async def flush(self) -> None:
        pass

    @abstractmethod
    async def aclose(self) -> None:
        pass

    @abstractmethod
    async def __anext__(self) -> SynthesisEvent:
        pass

    def __aiter__(self) -> "SynthesizeStream":
        return self


class TTS(ABC):
    def __init__(self, *, streaming_supported: bool) -> None:
        self._streaming_supported = streaming_supported

    @abstractmethod
    async def synthesize(self, *, text: str) -> SynthesizedAudio:
        pass

    def stream(self) -> SynthesizeStream:
        raise NotImplementedError(
            "streaming is not supported by this TTS, please use a different TTS or use a StreamAdapter"
        )

    @property
    def streaming_supported(self) -> bool:
        return self._streaming_supported



from typing import List
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional


@dataclass
class SegmentedSentence:
    text: str


class SentenceTokenizer(ABC):
    @abstractmethod
    def tokenize(
        self, *, text: str, language: Optional[str] = None
    ) -> List[SegmentedSentence]:
        pass

    @abstractmethod
    def stream(self, *, language: Optional[str] = None) -> "SentenceStream":
        pass


class SentenceStream(ABC):
    @abstractmethod
    def push_text(self, text: str) -> None:
        pass

    @abstractmethod
    async def flush(self) -> None:
        pass

    async def aclose(self) -> None:
        pass

    @abstractmethod
    async def __anext__(self) -> SegmentedSentence:
        pass

    def __aiter__(self) -> "SentenceStream":
        return self



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import signal
import uuid
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Tuple,
)
from urllib.parse import urlparse

import websockets

from livekit import api, protocol
from livekit.protocol import agent as proto_agent
from livekit.protocol import models as proto_models
from livekit.protocol.agent import JobType
from .job_request import JobRequest
from .job_context import JobContext
from .plugin import Plugin

MAX_RECONNECT_ATTEMPTS = 10
RECONNECT_INTERVAL = 5
ASSIGNMENT_TIMEOUT = 15

JobRequestHandler = Callable[["JobRequest"], Coroutine]


class AssignmentTimeoutError(Exception):
    """Worker timed out when joining the worker-pool"""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class JobCancelledError(Exception):
    """Job was cancelled by the server"""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class Worker:
    """A Worker is a client that connects to LiveKit Cloud (or a LiveKit server) and receives Agent jobs.
    For Job the Worker accepts, it will connect to the room and handle track subscriptions.
    """

    def __init__(
        self,
        request_handler: JobRequestHandler,
        *,
        worker_type: JobType.ValueType = JobType.JT_ROOM,
        event_loop: Optional[asyncio.AbstractEventLoop] = None,
        ws_url: Optional[str] = None,
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
    ) -> None:
        """
        Args:
            request_handler (JobRequestHandler): Callback that is triggered when a new Job is available.
            worker_type (JobType): What kind of jobs this worker can handle.
            event_loop (Optional[asyncio.AbstractEventLoop]): Optional asyncio event loop to use for this worker. Defaults to None.
            ws_url (str, optional): LiveKit websocket URL. Defaults to os.environ.get("LIVEKIT_URL", "http://localhost:7880").
            api_key (str, optional): LiveKit API Key. Defaults to os.environ.get("LIVEKIT_API_KEY", "").
            api_secret (str, optional): LiveKit API Secret. Defaults to os.environ.get("LIVEKIT_API_SECRET", "").
        """

        ws_url = ws_url or os.environ.get("LIVEKIT_URL") or "ws://localhost:7880"
        api_key = api_key or os.environ.get("LIVEKIT_API_KEY")
        api_secret = api_secret or os.environ.get("LIVEKIT_API_SECRET")

        if not api_key:
            raise ValueError(
                "No api key provided, set LIVEKIT_API_KEY or use the api-key parameter inside the CLI"
            )

        if not api_secret:
            raise ValueError(
                "No api secret provided, set LIVEKIT_API_SECRET or use the api-secret parameter inside the CLI"
            )

        self._set_url(ws_url)

        self._loop = event_loop or asyncio.get_event_loop()
        self._lock = asyncio.Lock()
        self._request_handler = request_handler
        self._wid = "W-" + str(uuid.uuid4())[:12]
        self._worker_type = worker_type
        self._api_key = api_key
        self._api_secret = api_secret
        self._running = False
        self._running_jobs: list["JobContext"] = []
        self._pending_jobs: Dict[str, asyncio.Future[proto_agent.JobAssignment]] = {}
        self._api = api.LiveKitAPI(ws_url, api_key, api_secret)

    def _set_url(self, ws_url: str) -> None:
        parse_res = urlparse(ws_url)
        scheme = parse_res.scheme
        if scheme.startswith("http"):
            scheme = scheme.replace("http", "ws")

        url = f"{scheme}://{parse_res.netloc}/{parse_res.path}"
        url = url.rstrip("/")

        self._agent_url = url + "/agent"
        self._rtc_url = url

    async def _connect(self) -> protocol.agent.RegisterWorkerResponse:
        join_jwt = (
            api.AccessToken(self._api_key, self._api_secret)
            .with_grants(api.VideoGrants(agent=True))
            .to_jwt()
        )

        req = protocol.agent.WorkerMessage()
        req.register.worker_id = self._wid
        req.register.type = self._worker_type

        headers = {"Authorization": f"Bearer {join_jwt}"}
        self._ws = await websockets.connect(
            self._agent_url, extra_headers=headers, close_timeout=0.150
        )
        await self._send(req)
        res = await self._recv()
        return res.register

    async def _send_availability(
        self, job_id: str, available: bool
    ) -> protocol.agent.JobAssignment:
        """Send availability to the server, and wait for assignment"""
        req = protocol.agent.WorkerMessage()
        req.availability.available = available
        req.availability.job_id = job_id

        f = asyncio.Future()
        self._pending_jobs[job_id] = f
        await self._send(req)

        try:
            return await asyncio.wait_for(f, ASSIGNMENT_TIMEOUT)
        except asyncio.TimeoutError as exc:
            raise AssignmentTimeoutError(
                f"assignment timeout for job {job_id}"
            ) from exc

    async def _send_job_status(
        self,
        job_id: str,
        status: protocol.agent.JobStatus.ValueType,
        error: str,
        user_data: str = "",
    ) -> None:
        req = protocol.agent.WorkerMessage()
        req.job_update.job_id = job_id
        req.job_update.status = status
        req.job_update.error = error
        req.job_update.user_data = user_data
        await self._ws.send(req.SerializeToString())

    def _simulate_job(
        self,
        room: proto_models.Room,
        participant: Optional[proto_models.ParticipantInfo],
    ):
        # TODO(theomonnom): the server could handle the JobSimulation like
        # we're doing with the SFU today
        job_id = "JR_" + str(uuid.uuid4())[:12]
        job_type = JobType.JT_ROOM if participant is None else JobType.JT_PUBLISHER
        job = proto_agent.Job(
            id=job_id, type=job_type, room=room, participant=participant
        )
        job = JobRequest(self, job, simulated=True)
        asyncio.ensure_future(self._handle_new_job(job), loop=self._loop)

    async def _recv(self) -> proto_agent.ServerMessage:
        message = await self._ws.recv()
        msg = protocol.agent.ServerMessage()
        msg.ParseFromString(bytes(message))  # type: ignore
        return msg

    async def _send(self, msg: protocol.agent.WorkerMessage) -> None:
        try:
            await self._ws.send(msg.SerializeToString())
        except websockets.exceptions.ConnectionClosed:
            # TODO: Implement JobStatus resuming after reconnection
            pass

    async def _handle_new_job(self, job: "JobRequest") -> None:
        """Execute the available callback, and automatically deny the job if the callback
        does not send an answer or raises an exception"""

        try:
            await self._request_handler(job)
        except Exception:
            logging.exception("request handler for job %s failed", job.id)
            return

        if not job._answered:
            logging.warning(
                "user did not answer availability for job %s, rejecting", job.id
            )
            await job.reject()

    async def _message_received(self, msg: protocol.agent.ServerMessage) -> None:
        logging.debug("received message: %s", msg)
        which = msg.WhichOneof("message")
        if which == "availability":
            # server is asking the worker if we are available for a job
            availability = msg.availability
            job = JobRequest(self, availability.job)
            asyncio.ensure_future(self._handle_new_job(job), loop=self._loop)
        elif which == "assignment":
            # server is assigning a job to the worker
            assignment = msg.assignment
            job_id = assignment.job.id
            f = self._pending_jobs.get(job_id)
            if f is None:
                logging.error("received assignment for unknown job %s", job_id)
                return

            f.set_result(assignment)
            del self._pending_jobs[job_id]

    async def _reconnect(self) -> bool:
        for i in range(MAX_RECONNECT_ATTEMPTS):
            try:
                reg = await self._connect()
                logging.info("worker successfully re-registered: %s", reg.worker_id)
                return True
            except Exception as e:
                logging.error("failed to reconnect, attempt %i: %s", i, e)
                await asyncio.sleep(RECONNECT_INTERVAL)

        return False

    async def _run(self) -> None:
        try:
            while True:
                try:
                    while True:
                        await self._message_received(await self._recv())
                except websockets.exceptions.ConnectionClosed as e:
                    if self._running:
                        logging.error("connection closed, trying to reconnect: %s", e)
                        if not await self._reconnect():
                            break
                except Exception as e:
                    logging.error("error while running worker: %s", e)
                    break
        finally:
            await asyncio.shield(self._shutdown())

    async def _shutdown(self) -> None:
        async with self._lock:
            if not self._running:
                return

            await self._ws.close()
            # Close all running jobs
            await asyncio.gather(*[job.disconnect() for job in self._running_jobs])
            self._running = False

    async def start(self) -> None:
        """Start the Worker"""
        async with self._lock:
            if self._running:
                raise Exception("worker is already running")

            await self._connect()  # initial connection
            self._running = True
            self._task = self._loop.create_task(self._run())

    async def shutdown(self) -> None:
        """Shut the Worker down."""
        async with self._lock:
            if not self._running:
                return

            self._task.cancel()
            await self._task

    @property
    def id(self) -> str:
        """Worker ID"""
        return self._wid

    @property
    def running(self) -> bool:
        """Whether the worker is running.
        Running is first set to True when the websocket connection is established and
        the Worker has been acknowledged by a LiveKit Server."""
        return self._running

    @property
    def api(self) -> api.LiveKitAPI:
        return self._api


def _run_worker(
    worker: Worker,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    started_cb: Optional[Callable[[Worker], Any]] = None,
) -> None:
    """Run the specified worker and handle graceful shutdown"""

    loop = loop or asyncio.get_event_loop()

    class GracefulShutdown(SystemExit):
        code = 1

    for sig in (signal.SIGINT, signal.SIGTERM):
        try:

            def _signal_handler():
                raise GracefulShutdown()

            loop.add_signal_handler(sig, _signal_handler)
        except NotImplementedError:
            pass

    async def _main_task(worker: Worker) -> None:
        try:
            await worker.start()
            if started_cb:
                started_cb(worker)

            logging.info(
                "worker started, press Ctrl+C to stop (worker id: %s)", worker.id
            )

            await worker._task
        except asyncio.CancelledError:
            pass
        finally:
            logging.info("shutting down worker %s", worker.id)
            await worker.shutdown()
            logging.info("worker %s shutdown", worker.id)

    main_task = loop.create_task(_main_task(worker))
    try:
        asyncio.set_event_loop(loop)
        loop.run_until_complete(main_task)
    except (GracefulShutdown, KeyboardInterrupt):
        logging.info("Graceful shutdown worker")
    finally:
        main_task.cancel()
        loop.run_until_complete(main_task)

        tasks = asyncio.all_tasks(loop)
        for task in tasks:
            task.cancel()

        loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()
        asyncio.set_event_loop(None)


def run_app(worker: Worker) -> None:
    """Run the CLI to interact with the worker"""

    import click

    @click.group()
    @click.option(
        "--log-level",
        default="INFO",
        type=click.Choice(
            ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
        ),
        help="Set the logging level",
    )
    @click.option(
        "--url",
        help="The websocket URL",
        default=worker._rtc_url,
    )
    @click.option("--api-key", help="The API key", default=worker._api_key)
    @click.option("--api-secret", help="The API secret", default=worker._api_secret)
    def cli(log_level: str, url: str, api_key: str, api_secret: str) -> None:
        logging.basicConfig(level=log_level)
        worker._set_url(url)
        worker._api_key = api_key
        worker._api_secret = api_secret

    @cli.command(help="Start the worker")
    def start() -> None:
        _run_worker(worker)

    @cli.command(help="Start a worker and simulate a job, useful for testing")
    @click.option("--room-name", help="The room name", required=True)
    @click.option("--identity", help="The participant identity")
    def simulate_job(room_name: str, identity: str) -> None:
        async def _pre_run() -> (
            Tuple[proto_models.Room, Optional[proto_models.ParticipantInfo]]
        ):
            lkapi = api.LiveKitAPI(worker._rtc_url, worker._api_key, worker._api_secret)

            try:
                room = await lkapi.room.create_room(
                    api.CreateRoomRequest(name=room_name)
                )

                participant = None
                if identity:
                    participant = await lkapi.room.get_participant(
                        api.RoomParticipantIdentity(room=room_name, identity=identity)
                    )

                return room, participant
            finally:
                await lkapi.aclose()

        room_info, participant = worker._loop.run_until_complete(_pre_run())
        logging.info(f"Simulating job for room {room_info.name} ({room_info.sid})")
        _run_worker(
            worker, started_cb=lambda _: worker._simulate_job(room_info, participant)
        )

    @cli.command(help="List used plugins")
    def plugins() -> None:
        for plugin in Plugin.registered_plugins:
            logging.info(plugin.title)

    @cli.command(help="Download required files of used plugins")
    @click.option("--exclude", help="Exclude plugins", multiple=True)
    def download_files(exclude: Tuple[str]) -> None:
        for plugin in Plugin.registered_plugins:
            if plugin.title in exclude:
                continue

            logging.info("Setup data for plugin %s", plugin.title)
            plugin.download_files()

    cli()



import logging
import asyncio
from .tts import (
    TTS,
    SynthesizeStream,
    SynthesisEvent,
    SynthesisEventType,
    SynthesizedAudio,
)
from ..tokenize import SentenceTokenizer, SentenceStream


class StreamAdapterWrapper(SynthesizeStream):
    def __init__(self, tts: TTS, sentence_stream: SentenceStream) -> None:
        super().__init__()
        self._closed = False
        self._tts = tts
        self._sentence_stream = sentence_stream
        self._queue = asyncio.Queue[str]()
        self._event_queue = asyncio.Queue[SynthesisEvent]()

        self._main_task = asyncio.create_task(self._run())

        def log_exception(task: asyncio.Task) -> None:
            if not task.cancelled() and task.exception():
                logging.error(f"google speech task failed: {task.exception()}")

        self._main_task.add_done_callback(log_exception)

    async def _run(self) -> None:
        while True:
            try:
                sentence = await self._sentence_stream.__anext__()
                audio = await self._tts.synthesize(text=sentence.text)
                self._event_queue.put_nowait(
                    SynthesisEvent(type=SynthesisEventType.AUDIO, audio=audio)
                )
            except asyncio.CancelledError:
                break

        self._closed = True

    def push_text(self, token: str) -> None:
        self._sentence_stream.push_text(token)

    async def flush(self) -> None:
        await self._sentence_stream.flush()

    async def aclose(self) -> None:
        self._main_task.cancel()
        try:
            await self._main_task
        except asyncio.CancelledError:
            pass

    async def __anext__(self) -> SynthesisEvent:
        if self._closed and self._event_queue.empty():
            raise StopAsyncIteration

        return await self._event_queue.get()


class StreamAdapter(TTS):
    def __init__(self, tts: TTS, tokenizer: SentenceTokenizer) -> None:
        super().__init__(streaming_supported=True)
        self._tts = tts
        self._tokenizer = tokenizer

    async def synthesize(self, *, text: str) -> SynthesizedAudio:
        return await self._tts.synthesize(text=text)

    def stream(self) -> SynthesizeStream:
        return StreamAdapterWrapper(self._tts, self._tokenizer.stream())
